import sys
from datetime import datetime
from functools import partial
from types import SimpleNamespace

from rich import print as rprint

from skyplane.utils.definitions import is_gateway_env

log_file = None


def open_log_file(filename):
    global log_file
    log_file = open(filename, "a")


def log(msg, LEVEL="INFO", color="white", write_to_file=True, write_to_stderr=True, *args, **kwargs):
    if args or kwargs:
        msg = msg.format(*args, **kwargs)
    level_prefix = ("[" + LEVEL.upper() + "]").ljust(7)
    time = datetime.now().strftime("%H:%M:%S")
    if write_to_file and log_file:
        log_file.write(f"{time} {level_prefix} {msg}\n")
        log_file.flush()
    if write_to_stderr:
        rprint(f"{time} {level_prefix} [{color}]{msg}[/]", flush=True, file=sys.stderr)


debug = partial(log, LEVEL="DEBUG", color="cyan")
info = partial(log, LEVEL="INFO", color="white")
warn = partial(log, LEVEL="WARN", color="yellow")
warning = partial(log, LEVEL="WARN", color="yellow")
error = partial(log, LEVEL="ERROR", color="red")


def exception(msg, print_traceback=True, write_to_file=False, *args, **kwargs):
    error(f"Exception: {msg}", write_to_file=write_to_file, *args, **kwargs)
    if print_traceback:
        import traceback

        if write_to_file:
            traceback.print_exc(file=log_file)
        else:
            traceback.print_exc()


# define fs object to log to disk log
fs = SimpleNamespace(
    debug=partial(log, LEVEL="DEBUG", color="cyan", write_to_file=True, write_to_stderr=is_gateway_env),
    info=partial(log, LEVEL="INFO", color="white", write_to_file=True, write_to_stderr=is_gateway_env),
    warn=partial(log, LEVEL="WARN", color="yellow", write_to_file=True, write_to_stderr=is_gateway_env),
    warning=partial(log, LEVEL="WARN", color="yellow", write_to_file=True, write_to_stderr=is_gateway_env),
    error=partial(log, LEVEL="ERROR", color="red", write_to_file=True, write_to_stderr=is_gateway_env),
    exception=partial(exception, write_to_file=True, write_to_stderr=is_gateway_env),
    log=partial(log, write_to_file=True, write_to_stderr=is_gateway_env),
)



from importlib.resources import path

from skyplane import compute
from skyplane.planner.solver import ThroughputProblem
from skyplane.planner.topology import ReplicationTopology


class Planner:
    def __init__(self, src_provider: str, src_region, dst_provider: str, dst_region: str):
        self.src_provider = src_provider
        self.src_region = src_region
        self.dst_provider = dst_provider
        self.dst_region = dst_region

    def plan(self) -> ReplicationTopology:
        raise NotImplementedError


class DirectPlanner(Planner):
    def __init__(self, src_provider: str, src_region, dst_provider: str, dst_region: str, n_instances: int, n_connections: int):
        self.n_instances = n_instances
        self.n_connections = n_connections
        super().__init__(src_provider, src_region, dst_provider, dst_region)

    def plan(self) -> ReplicationTopology:
        src_region_tag = f"{self.src_provider}:{self.src_region}"
        dst_region_tag = f"{self.dst_provider}:{self.dst_region}"
        if src_region_tag == dst_region_tag:  # intra-region transfer w/o solver
            topo = ReplicationTopology()
            for i in range(self.n_instances):
                topo.add_objstore_instance_edge(src_region_tag, src_region_tag, i)
                topo.add_instance_objstore_edge(src_region_tag, i, src_region_tag)
            topo.cost_per_gb = 0
            return topo
        else:  # inter-region transfer w/ solver
            topo = ReplicationTopology()
            for i in range(self.n_instances):
                topo.add_objstore_instance_edge(src_region_tag, src_region_tag, i)
                topo.add_instance_instance_edge(src_region_tag, i, dst_region_tag, i, self.n_connections)
                topo.add_instance_objstore_edge(dst_region_tag, i, dst_region_tag)
            topo.cost_per_gb = compute.CloudProvider.get_transfer_cost(src_region_tag, dst_region_tag)
            return topo


class ILPSolverPlanner(Planner):
    def __init__(
        self,
        src_provider: str,
        src_region,
        dst_provider: str,
        dst_region: str,
        max_instances: int,
        max_connections: int,
        required_throughput_gbits: float,
    ):
        self.max_instances = max_instances
        self.max_connections = max_connections
        self.solver_required_throughput_gbits = required_throughput_gbits
        super().__init__(src_provider, src_region, dst_provider, dst_region)

    def plan(self) -> ReplicationTopology:
        from skyplane.planner.solver_ilp import ThroughputSolverILP

        problem = ThroughputProblem(
            src=f"{self.src_provider}:{self.src_region}",
            dst=f"{self.dst_provider}:{self.dst_region}",
            required_throughput_gbits=self.solver_required_throughput_gbits,
            gbyte_to_transfer=1,
            instance_limit=self.max_instances,
        )

        with path("skyplane.data", "throughput.csv") as solver_throughput_grid:
            tput = ThroughputSolverILP(solver_throughput_grid)
        solution = tput.solve_min_cost(problem, solver=ThroughputSolverILP.choose_solver(), save_lp_path=None)
        if not solution.is_feasible:
            raise ValueError("ILP solver failed to find a solution, try solving with fewer constraints")
        topo, _ = tput.to_replication_topology(solution)
        return topo


class RONSolverPlanner(Planner):
    def __init__(
        self,
        src_provider: str,
        src_region,
        dst_provider: str,
        dst_region: str,
        max_instances: int,
        max_connections: int,
        required_throughput_gbits: float,
    ):
        self.max_instances = max_instances
        self.max_connections = max_connections
        self.solver_required_throughput_gbits = required_throughput_gbits
        super().__init__(src_provider, src_region, dst_provider, dst_region)

    def plan(self) -> ReplicationTopology:
        from skyplane.planner.solver_ron import ThroughputSolverRON

        problem = ThroughputProblem(
            src=self.src_region,
            dst=self.dst_region,
            required_throughput_gbits=self.solver_required_throughput_gbits,
            gbyte_to_transfer=1,
            instance_limit=self.max_instances,
        )

        with path("skyplane.data", "throughput.csv") as solver_throughput_grid:
            tput = ThroughputSolverRON(solver_throughput_grid)
        solution = tput.solve(problem)
        topo, _ = tput.to_replication_topology(solution)
        return topo



import json
import os
import threading
from collections import defaultdict, Counter
from functools import partial

import nacl.secret
import nacl.utils
import urllib3
from typing import TYPE_CHECKING, Dict, List, Optional

from skyplane import compute
from skyplane.api.tracker import TransferProgressTracker, TransferHook
from skyplane.api.transfer_job import CopyJob, SyncJob, TransferJob
from skyplane.api.config import TransferConfig
from skyplane.planner.topology import ReplicationTopology, ReplicationTopologyGateway
from skyplane.utils import logger
from skyplane.utils.definitions import gateway_docker_image
from skyplane.utils.fn import PathLike, do_parallel

if TYPE_CHECKING:
    from skyplane.api.provisioner import Provisioner


class DataplaneAutoDeprovision:
    def __init__(self, dataplane: "Dataplane"):
        self.dataplane = dataplane

    def __enter__(self):
        return self.dataplane

    def __exit__(self, exc_type, exc_value, exc_tb):
        logger.fs.warning("Deprovisioning dataplane")
        self.dataplane.deprovision()


class Dataplane:
    """A Dataplane represents a concrete Skyplane network, including topology and VMs."""

    def __init__(
        self,
        clientid: str,
        topology: ReplicationTopology,
        provisioner: "Provisioner",
        transfer_config: TransferConfig,
    ):
        """
        :param clientid: the uuid of the local host to create the dataplane
        :type clientid: str
        :param topology: the calculated topology during the transfer
        :type topology: ReplicationTopology
        :param provisioner: the provisioner to launch the VMs
        :type provisioner: Provisioner
        :param transfer_config: the configuration during the transfer
        :type transfer_config: TransferConfig
        """
        self.clientid = clientid
        self.topology = topology
        self.src_region_tag = self.topology.source_region()
        self.dst_region_tag = self.topology.sink_region()
        regions = Counter([node.region for node in self.topology.gateway_nodes])
        self.max_instances = int(regions[max(regions, key=regions.get)])
        self.provisioner = provisioner
        self.transfer_config = transfer_config
        self.http_pool = urllib3.PoolManager(retries=urllib3.Retry(total=3))
        self.provisioning_lock = threading.Lock()
        self.provisioned = False

        # pending tracker tasks
        self.jobs_to_dispatch: List[TransferJob] = []
        self.pending_transfers: List[TransferProgressTracker] = []
        self.bound_nodes: Dict[ReplicationTopologyGateway, compute.Server] = {}

    def provision(
        self,
        allow_firewall: bool = True,
        gateway_docker_image: str = os.environ.get("SKYPLANE_DOCKER_IMAGE", gateway_docker_image()),
        gateway_log_dir: Optional[PathLike] = None,
        authorize_ssh_pub_key: Optional[str] = None,
        max_jobs: int = 16,
        spinner: bool = False,
    ):
        """
        Provision the transfer gateways.

        :param allow_firewall: whether to apply firewall rules in the gatweway network (default: True)
        :type allow_firewall: bool
        :param gateway_docker_image: Docker image token in github
        :type gateway_docker_image: str
        :param gateway_log_dir: path to the log directory in the remote gatweways
        :type gateway_log_dir: PathLike
        :param authorize_ssh_pub_key: authorization ssh key to the remote gateways
        :type authorize_ssh_pub_key: str
        :param max_jobs: maximum number of provision jobs to launch concurrently (default: 16)
        :type max_jobs: int
        :param spinner: whether to show the spinner during the job (default: False)
        :type spinner: bool
        """
        with self.provisioning_lock:
            if self.provisioned:
                logger.error("Cannot provision dataplane, already provisioned!")
                return
            aws_nodes_to_provision = list(n.region.split(":")[0] for n in self.topology.nodes if n.region.startswith("aws:"))
            azure_nodes_to_provision = list(n.region.split(":")[0] for n in self.topology.nodes if n.region.startswith("azure:"))
            gcp_nodes_to_provision = list(n.region.split(":")[0] for n in self.topology.nodes if n.region.startswith("gcp:"))

            # create VMs from the topology
            for node in self.topology.gateway_nodes:
                cloud_provider, region = node.region.split(":")
                self.provisioner.add_task(
                    cloud_provider=cloud_provider,
                    region=region,
                    vm_type=getattr(self.transfer_config, f"{cloud_provider}_instance_class"),
                    spot=getattr(self.transfer_config, f"{cloud_provider}_use_spot_instances"),
                    autoterminate_minutes=self.transfer_config.autoterminate_minutes,
                )

            # initialize clouds
            self.provisioner.init_global(
                aws=len(aws_nodes_to_provision) > 0,
                azure=len(azure_nodes_to_provision) > 0,
                gcp=len(gcp_nodes_to_provision) > 0,
            )

            # provision VMs
            uuids = self.provisioner.provision(
                authorize_firewall=allow_firewall,
                max_jobs=max_jobs,
                spinner=spinner,
            )

            # bind VMs to nodes
            servers = [self.provisioner.get_node(u) for u in uuids]
            servers_by_region = defaultdict(list)
            for s in servers:
                servers_by_region[s.region_tag].append(s)
            for node in self.topology.gateway_nodes:
                instance = servers_by_region[node.region].pop()
                self.bound_nodes[node] = instance
            logger.fs.debug(f"[Dataplane.provision] bound_nodes = {self.bound_nodes}")
            gateway_bound_nodes = self.bound_nodes.copy()

            # start gateways
            self.provisioned = True

        def _start_gateway(
            gateway_node: ReplicationTopologyGateway,
            gateway_server: compute.Server,
        ):
            # map outgoing ports
            setup_args = {}
            for n, v in self.topology.get_outgoing_paths(gateway_node).items():
                if isinstance(n, ReplicationTopologyGateway):
                    # use private ips for gcp to gcp connection
                    src_provider, dst_provider = gateway_node.region.split(":")[0], n.region.split(":")[0]
                    if src_provider == dst_provider and src_provider == "gcp":
                        setup_args[self.bound_nodes[n].private_ip()] = v
                    else:
                        setup_args[self.bound_nodes[n].public_ip()] = v
            am_source = gateway_node in self.topology.source_instances()
            am_sink = gateway_node in self.topology.sink_instances()
            logger.fs.debug(f"[Dataplane._start_gateway] Setup args for {gateway_node}: {setup_args}")

            # start gateway
            if gateway_log_dir:
                gateway_server.init_log_files(gateway_log_dir)
            if authorize_ssh_pub_key:
                gateway_server.copy_public_key(authorize_ssh_pub_key)
            gateway_server.start_gateway(
                setup_args,
                gateway_docker_image=gateway_docker_image,
                e2ee_key_bytes=e2ee_key_bytes if (self.transfer_config.use_e2ee and (am_source or am_sink)) else None,
                use_bbr=self.transfer_config.use_bbr,
                use_compression=self.transfer_config.use_compression,
                use_socket_tls=self.transfer_config.use_socket_tls,
            )

        # todo: move server.py:start_gateway here
        logger.fs.info(f"Using docker image {gateway_docker_image}")
        e2ee_key_bytes = nacl.utils.random(nacl.secret.SecretBox.KEY_SIZE)

        jobs = []
        for node, server in gateway_bound_nodes.items():
            jobs.append(partial(_start_gateway, node, server))
        logger.fs.debug(f"[Dataplane.provision] Starting gateways on {len(jobs)} servers")
        do_parallel(lambda fn: fn(), jobs, n=-1, spinner=spinner, spinner_persist=spinner, desc="Starting gateway container on VMs")

    def deprovision(self, max_jobs: int = 64, spinner: bool = False):
        """
        Deprovision the remote gateways

        :param max_jobs: maximum number of jobs to deprovision the remote gateways (default: 64)
        :type max_jobs: int
        :param spinner: Whether to show the spinner during the job (default: False)
        :type spinner: bool
        """
        with self.provisioning_lock:
            if not self.provisioned:
                logger.fs.warning("Attempting to deprovision dataplane that is not provisioned, this may be from auto_deprovision.")
            # wait for tracker tasks
            try:
                for task in self.pending_transfers:
                    logger.fs.warning(f"Before deprovisioning, waiting for jobs to finish: {list(task.jobs.keys())}")
                    task.join()
            except KeyboardInterrupt:
                logger.warning("Interrupted while waiting for transfers to finish, deprovisioning anyway.")
                raise
            finally:
                self.provisioner.deprovision(
                    max_jobs=max_jobs,
                    spinner=spinner,
                )
                self.provisioned = False

    def check_error_logs(self) -> Dict[str, List[str]]:
        """Get the error log from remote gateways if there is any error."""

        def get_error_logs(args):
            _, instance = args
            reply = self.http_pool.request("GET", f"{instance.gateway_api_url}/api/v1/errors")
            if reply.status != 200:
                raise Exception(f"Failed to get error logs from gateway instance {instance.instance_name()}: {reply.data.decode('utf-8')}")
            return json.loads(reply.data.decode("utf-8"))["errors"]

        errors: Dict[str, List[str]] = {}
        for (_, instance), result in do_parallel(get_error_logs, self.bound_nodes.items(), n=-1):
            errors[instance] = result
        return errors

    def auto_deprovision(self) -> DataplaneAutoDeprovision:
        """Returns a context manager that will automatically call deprovision upon exit."""
        return DataplaneAutoDeprovision(self)

    def source_gateways(self) -> List[compute.Server]:
        """Returns a list of source gateway nodes"""
        return [self.bound_nodes[n] for n in self.topology.source_instances()] if self.provisioned else []

    def sink_gateways(self) -> List[compute.Server]:
        """Returns a list of sink gateway nodes"""
        return [self.bound_nodes[n] for n in self.topology.sink_instances()] if self.provisioned else []

    def queue_copy(
        self,
        src: str,
        dst: str,
        recursive: bool = False,
    ) -> str:
        """
        Add a copy job to job list.
        Return the uuid of the job.

        :param src: source prefix to copy from
        :type src: str
        :param dst: the destination of the transfer
        :type dst: str
        :param recursive: if true, will copy objects at folder prefix recursively (default: False)
        :type recursive: bool
        """
        job = CopyJob(src, dst, recursive, requester_pays=self.transfer_config.requester_pays)
        logger.fs.debug(f"[SkyplaneClient] Queued copy job {job}")
        self.jobs_to_dispatch.append(job)
        return job.uuid

    def queue_sync(
        self,
        src: str,
        dst: str,
    ) -> str:
        """
        Add a sync job to job list.
        Return the uuid of the job.

        :param src: Source prefix to copy from
        :type src: str
        :param dst: The destination of the transfer
        :type dst: str
        :param recursive: If true, will copy objects at folder prefix recursively (default: False)
        :type recursive: bool
        """
        job = SyncJob(src, dst, recursive=True, requester_pays=self.transfer_config.requester_pays)
        logger.fs.debug(f"[SkyplaneClient] Queued sync job {job}")
        self.jobs_to_dispatch.append(job)
        return job.uuid

    def run_async(self, hooks: Optional[TransferHook] = None) -> TransferProgressTracker:
        """Start the transfer asynchronously. The main thread will not be blocked.

        :param hooks: Tracks the status of the transfer
        :type hooks: TransferHook
        """
        if not self.provisioned:
            logger.error("Dataplane must be pre-provisioned. Call dataplane.provision() before starting a transfer")
        tracker = TransferProgressTracker(self, self.jobs_to_dispatch, self.transfer_config, hooks)
        self.pending_transfers.append(tracker)
        tracker.start()
        logger.fs.info(f"[SkyplaneClient] Started async transfer with {len(self.jobs_to_dispatch)} jobs")
        self.jobs_to_dispatch = []
        return tracker

    def run(self, hooks: Optional[TransferHook] = None):
        """Start the transfer in the main thread. Wait until the transfer is complete.

        :param hooks: Tracks the status of the transfer
        :type hooks: TransferHook
        """
        tracker = self.run_async(hooks)
        logger.fs.debug(f"[SkyplaneClient] Waiting for transfer to complete")
        tracker.join()
